#ifndef PROFINET_SNIFFER_HPP
#define PROFINET_SNIFFER_HPP

#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <functional>
#include <system_error>
#include <memory>
#include <thread>
#include <string>
#include <mutex>
#include <unordered_map>
#include <uv.h>

#ifdef PROFINET_DEBUG
#define _PROFINET_DEBUG \
    if (1) std::cout << "cybus_profinet::" << __func__ << "\t"
#else
#define _PROFINET_DEBUG \
    if (0) std::cout << "cybus_profinet::" << __func__ << "\t"
#endif

namespace cybus_profinet {

  struct DeviceInfo {
    std::string device_name;
    std::uint16_t vendor_id;
    std::string ip;
    std::string mac;
    std::uint16_t device_id;
    std::string device_type;
  };

  struct ProfinetValue {
    bool is_error;
    std::vector<uint8_t> buffer;
    int32_t slot;
    int32_t subslot;
    int32_t property_index;
    std::string data_type;
  };

  class ProfinetSniffer {

    // PIMPL (pointer to implementation) design pattern
    class Impl;
    std::shared_ptr<Impl> _impl;

    typedef std::function<void (ProfinetValue)> NodeCallback;

    typedef std::unordered_map<size_t, NodeCallback> NodeCallbacks;
    static NodeCallbacks awaited_callbacks;
    static size_t invoke_id;

    typedef std::pair<size_t, ProfinetValue> AsyncData;
    static std::vector<AsyncData> _data_queue;
    static std::mutex _data_queue_mutex;
    static uv_async_t async;
    static std::hash<std::string> _hasher;

  public:

    ProfinetSniffer(const std::string& interface, const std::string& gsd_path,
      const std::string& config_path);

    ~ProfinetSniffer();

    void start();

    void stop();

    void reparseGsdFiles();

    std::vector<DeviceInfo> getDevices();

    void subscribe(const std::string& device_name, int32_t api, int32_t slot,
    int32_t sub_slot, int32_t property_index, const NodeCallback& callback);

    void subscribe(const std::string& device_name, int32_t api, int32_t slot,
    int32_t sub_slot, const std::string& property_name,
    const NodeCallback& callback);

    void unsubscribe(const std::string& device_name, int32_t api, int32_t slot,
      int32_t sub_slot, int32_t property_index);

    void unsubscribe(const std::string& device_name, int32_t api, int32_t slot,
      int32_t sub_slot, const std::string& property_name);

  private:

    void handle_api_callback(bool is_error, char* buffer,
      int32_t length, int32_t type, int32_t invoke_id);

    static void trigger_node_callback(uv_async_t* handle);

    template <typename A, typename B, typename C, typename D, typename E>
    size_t generate_id(const A& a, const B& b, const C& c, const D& d, const E& e) {
      std::ostringstream os;
      os << a << b << c << d << e;
      return _hasher(os.str());
    }
  };
}

#endif
